/* change the brand and text color */                                           
.navbar-custom .navbar-brand,                                                   
.navbar-custom .navbar-text {                                                   
    color: rgba(255,255,255,.7);                                                
}                                                                               
/* change the link color */                                                     
.navbar-custom .navbar-nav .nav-link {                                          
    color: rgba(255,255,255,.7);                                                
}                                                                               
/* change the color of active or hovered links */                               
.navbar-custom .nav-item.active .nav-link,                                      
.navbar-custom .nav-item:focus .nav-link,                                       
.navbar-custom .nav-item:hover .nav-link {                                      
    color: rgba(255,255,255,.9);                                                
} 
